"""
Written by Matteo Dunnhofer - 2019

Class that defines the A3CTD tracker
"""
import sys
sys.path.insert(0, '..')
import os
import copy
import numpy as np
import torch
from config import Configuration
from model.ACModel import ActorCriticModel
from data.DataTransformer import DataTransformer
import utils as ut
from demotrackers.SiamFC.siamfc import TrackerSiamFC


class A3CTD(object):

    def __init__(self, save_img_preds=False):
        super(A3CTD, self).__init__()

        self.cfg = Configuration()
        self.data_transformer = DataTransformer(self.cfg)

        self.save_img_preds = save_img_preds

        if self.cfg.USE_GPU:
            self.gpu_id = self.cfg.GPU_ID
            self.device = torch.device('cuda', self.gpu_id)
        else:
            self.device = torch.device('cpu')

        self.model = ActorCriticModel(self.cfg).to(self.device)
        self.model.load_state_dict(torch.load(self.cfg.A3CTD_PATH))
        self.model.eval()

        self.expert_tracker = TrackerSiamFC(net_path=self.cfg.EXPERT_PATH, gpu_id=self.gpu_id)

    def init(self, image, box):
        # initialize expert tracker
        self.expert_tracker.init(image, box)

        image = image.convert('RGB')

        self.curr_bb = copy.deepcopy(box)
        self.expert_curr_bb = copy.deepcopy(box)

        self.prev_img = copy.deepcopy(image)

        self.model_state = copy.deepcopy(self.model.init_state(self.device))
        self.expert_model_state = copy.deepcopy(self.model.init_state(self.device))

        self.step = 0

        if self.save_img_preds:
            img = ut.draw_box(image, self.curr_bb, color='magenta')
            img = ut.draw_text(img, str(self.step), color='yellow')
            img.save(os.path.join(self.OUT_IMG_PATH, str(self.step) + '.jpg'))


    def update(self, image):
        # get expert prediction
        expert_pred = self.expert_tracker.update(image)

        image = image.convert('RGB')

        self.step += 1

        #if self.step % 20 == 0:
        #    self.model_state = copy.deepcopy(self.model.init_state(self.device))
        #    self.expert_model_state = copy.deepcopy(self.model.init_state(self.device))

        # tracker prediction
        bb = ut.get_crop_bb(copy.deepcopy(self.curr_bb), image.size[0], image.size[1], 1.5)

        state1 = self.data_transformer.preprocess_img(self.prev_img, bb).to(self.device).unsqueeze(0)
        state2 = self.data_transformer.preprocess_img(image, bb).to(self.device).unsqueeze(0)

        # get a3c tracker bounding box estimate
        policy_mu, value, self.model_state = self.model(state1, state2, self.model_state, self.device)
        mu = torch.clamp(policy_mu, -1.0, 1.0)
        action = self.denorm_action(mu.data.cpu().numpy()[0], self.curr_bb)

        # expert evaluation
        bb = ut.get_crop_bb(copy.deepcopy(self.expert_curr_bb), image.size[0], image.size[1], 1.5)

        state1 = self.data_transformer.preprocess_img(self.prev_img, bb).to(self.device).unsqueeze(0)
        state2 = self.data_transformer.preprocess_img(image, bb).to(self.device).unsqueeze(0)
        # state = Variable(torch.cat(crops).to(self.device))

        expert_policy_mu, expert_value, self.expert_model_state = self.model(state1, state2, self.expert_model_state, self.device)
        expert_mu = torch.clamp(expert_policy_mu, -1.0, 1.0)
        expert_action = self.denorm_action(expert_mu.data.cpu().numpy()[0], self.expert_curr_bb)

        #self.curr_bb = copy.deepcopy(action)
        self.expert_curr_bb = copy.deepcopy(expert_pred)

        self.prev_img = copy.deepcopy(image)

        # evaluate the quality of the a3ctracker and the expert and chhose one of the two
        if value.data.cpu().numpy()[0, 0] >= expert_value.data.cpu().numpy()[0, 0]:
            self.curr_bb = copy.deepcopy(action)
            #return self.cunp.array(self.curr_bb)
        else:
            self.curr_bb = copy.deepcopy(expert_pred)
            #self.model_state = (self.expert_model_state[0].clone(), self.expert_model_state[1].clone())
            #return np.array(expert_pred)

        if self.save_img_preds:
            img = ut.draw_box(image, self.expert_curr_bb, color='yellow')

            img = ut.draw_box(img, self.curr_bb, color='green')

            img = ut.draw_text(img, str(self.step), color='yellow')

            img.save(os.path.join(self.OUT_IMG_PATH, str(self.step) + '.jpg'))

        return self.curr_bb


    def denorm_action(self, action, old_bb):
        # from shift to bounding box coordinates
        norm_coeff = 1.0
        new_x = old_bb[0] + (norm_coeff * action[0] * old_bb[2])
        new_y = old_bb[1] + (norm_coeff * action[1] * old_bb[3])
        new_w = old_bb[2] + (norm_coeff * action[2] * old_bb[2])
        new_h = old_bb[3] + (norm_coeff * action[3] * old_bb[3])

        return [new_x, new_y, new_w, new_h]
